package de.unikoblenz.fgbks.core.dmn.verification.verifier.impl;

import static de.unikoblenz.fgbks.core.dmn.domain.vdmn.utils.VDmnFunctions.templateDecision;

import de.unikoblenz.fgbks.base.domain.Name;
import de.unikoblenz.fgbks.base.utils.UniqueIdGenerator;
import de.unikoblenz.fgbks.base.utils.boundary.Boundary;
import de.unikoblenz.fgbks.base.utils.boundary.bicreater.BoundaryBiCreaterType;
import de.unikoblenz.fgbks.base.utils.boundary.checker.BoundaryCheckType;
import de.unikoblenz.fgbks.core.dmn.domain.common.RowNumber;
import de.unikoblenz.fgbks.core.dmn.domain.ids.InputEntryId;
import de.unikoblenz.fgbks.core.dmn.domain.ids.RuleId;
import de.unikoblenz.fgbks.core.dmn.domain.vdmn.VDmnDecision;
import de.unikoblenz.fgbks.core.dmn.domain.vdmn.VDmnDecisionTable;
import de.unikoblenz.fgbks.core.dmn.domain.vdmn.VDmnInputColumn;
import de.unikoblenz.fgbks.core.dmn.domain.vdmn.VDmnInputValue;
import de.unikoblenz.fgbks.core.dmn.domain.vdmn.VDmnRule;
import de.unikoblenz.fgbks.core.dmn.domain.vdmn.impl.VDmnInputValueChangeableImpl;
import de.unikoblenz.fgbks.core.dmn.domain.vdmn.impl.VDmnInputValueImpl;
import de.unikoblenz.fgbks.core.dmn.domain.vdmn.impl.VDmnRuleChangeableImpl;
import de.unikoblenz.fgbks.core.dmn.domain.vdmn.impl.VDmnRuleImpl;
import de.unikoblenz.fgbks.core.dmn.domain.vdmn.impl.VDmnRuleImpl.Builder;
import de.unikoblenz.fgbks.core.dmn.verification.result.VerificationResultEntry.VerificationClassification;
import de.unikoblenz.fgbks.core.dmn.verification.result.VerificationResultEntryElement;
import de.unikoblenz.fgbks.core.dmn.verification.result.actions.Action;
import de.unikoblenz.fgbks.core.dmn.verification.result.actions.ActionScope;
import de.unikoblenz.fgbks.core.dmn.verification.result.actions.ActionType;
import de.unikoblenz.fgbks.core.dmn.verification.result.actions.VerificationFix;
import de.unikoblenz.fgbks.core.dmn.verification.verifier.AbstractVerifier;
import de.unikoblenz.fgbks.core.dmn.verification.verifier.DmnVerifier;
import de.unikoblenz.fgbks.core.dmn.verification.verifier.types.MissingRuleVerification;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;

@DmnVerifier(verifierType = MissingRuleVerification.class)
public class MissingRuleVerifier extends AbstractVerifier {

  private static Name FIX_NAME = new Name("Add rule");
  public static final long INITIAL_ROW_NUMER = 9999L;

  @Override
  protected void doVerification() {
    for (VDmnDecision decision : dmnObjectContainer.getVDmnDefinition().getDmnDecisions()) {
      findMissingRules(decision.getDmnDecisionTable());
    }
  }

  private void findMissingRules(VDmnDecisionTable dmnDecisionTable) {
    List<VDmnInputColumn> inputs = dmnDecisionTable.getInputColumns();
    List<VDmnRuleChangeableImpl> missingRules =
        checkForMissingRules(inputs, dmnDecisionTable.getRules());
    // add errors for missing intervals
    for (VDmnRule missingRule : missingRules) {
      StringBuilder sb = new StringBuilder();
      sb.append(templateDecision(dmnDecisionTable.getDmnDecision()));
      sb.append("The following rule is not defined: { ");
      sb.append(
          missingRule.getDmnInputValues().stream()
              .map(v -> v.getBoundary().getParsedText())
              .map(s -> s.isEmpty() ? "-" : s)
              .collect(Collectors.joining(" / ")));
      sb.append(" }");
      Action.Builder fixActionBuilder =
          Action.getBuilder().withActionScope(ActionScope.RULE).withActionType(ActionType.CREATE);
      for (VDmnInputValue v : missingRule.getDmnInputValues()) {
        fixActionBuilder.addValue(
            v.getDmnInputColumn().getInputId().getValue(), v.getBoundary().getParsedText());
      }
      vreFactory
          .addVerificationFix(VerificationFix.SHOW_DECISION)
          .addVerificationFix(
              VerificationFix.getBuilder()
                  .withFixName(FIX_NAME)
                  .addAction(fixActionBuilder.build())
                  .build())
          .addElement(VerificationResultEntryElement.create(dmnDecisionTable))
          .addToEntry(VerificationClassification.WARNING, sb.toString());
    }
  }

  private List<VDmnRuleChangeableImpl> checkForMissingRules(
      List<VDmnInputColumn> inputs, List<VDmnRule> rules) {
    // check, if missingIntervals are empty
    List<VDmnRuleChangeableImpl> missingRules = new LinkedList<>();
    missingRules.add(constructPseudoMissingRule(inputs));
    // Check new Missing Rules for each rule
    for (VDmnRule rule : rules) {
      checkForMissingRules(inputs, rule, missingRules);
    }
    mergeRules(missingRules);
    return missingRules;
  }

  private VDmnRuleChangeableImpl constructPseudoMissingRule(List<VDmnInputColumn> inputs) {
    Builder b =
        VDmnRuleImpl.getBuilder()
            .withRuleId(new RuleId("TODO")) // TODO
            .withDmnDecisionTable(inputs.get(0).getDmnDecisionTable())
            .withRowNumber(new RowNumber(INITIAL_ROW_NUMER));

    for (VDmnInputColumn col : inputs) {
      b.addInputValue(
          VDmnInputValueImpl.getBuilder()
              .withInputEntryId(new InputEntryId("TODO")) // TODO
              .withText("")
              .withDmnInputColumn(col)
              .withDmnRuleFromBuilder(b)
              .build());
    }
    return new VDmnRuleChangeableImpl(b.build());
  }

  private void checkForMissingRules(
      List<VDmnInputColumn> inputs,
      VDmnRule currentRule,
      List<VDmnRuleChangeableImpl> missingRules) {
    // search the missing rules, which are in contact with the current rule
    List<VDmnRuleChangeableImpl> inContactRules = new LinkedList<>();
    List<VDmnRuleChangeableImpl> inEqualRules = new LinkedList<>();
    for (VDmnRuleChangeableImpl missingRule : missingRules) {
      if (isInContact(missingRule, currentRule)) {
        // dont add equal rules
        if (isEqualRule(missingRule, currentRule)) {
          inEqualRules.add(missingRule);
        } else {
          inContactRules.add(missingRule);
        }
      }
    }
    // remove found rules from current missingRules
    missingRules.removeAll(inEqualRules);
    missingRules.removeAll(inContactRules);
    // Build new missing rules from in contact rules
    List<VDmnRuleChangeableImpl> newMissingRules = new LinkedList<>();
    for (VDmnRuleChangeableImpl inContactRule : inContactRules) {
      newMissingRules.addAll(constructNewMissingRules(inputs, inContactRule, currentRule));
    }
    // mergeRules(newMissingRules);
    missingRules.addAll(newMissingRules);
    // mergeRules(missingRules);
  }

  private boolean isEqualRule(VDmnRule missingRule, VDmnRule currentRule) {
    List<VDmnInputValue> inValsMissingRule = missingRule.getDmnInputValues();
    List<VDmnInputValue> inValsCurrentRule = currentRule.getDmnInputValues();
    for (int i = 0; i < inValsMissingRule.size(); i++) {
      if (!inValsMissingRule
          .get(i)
          .getBoundary()
          .checkWith(BoundaryCheckType.IS_EQUAL, inValsCurrentRule.get(i).getBoundary())) {
        return false;
      }
    }
    return true;
  }

  private List<VDmnRuleChangeableImpl> constructNewMissingRules(
      List<VDmnInputColumn> inputs, VDmnRuleChangeableImpl missingRule, VDmnRule currentRule) {
    List<VDmnRuleChangeableImpl> newMissingRules = new LinkedList<>();
    List<VDmnInputValue> missingInValues = new ArrayList<>(missingRule.getDmnInputValues());
    List<VDmnInputValue> currentRuleValues = new ArrayList<>(currentRule.getDmnInputValues());
    for (int i = 0; i < missingInValues.size(); i++) {
      // change bound for i-1
      if (i != 0) {
        VDmnInputValue missingRuleValue = missingInValues.get(i - 1);
        Boundary missingRuleBoundary = missingRuleValue.getBoundary();
        Boundary valueBoundary = currentRuleValues.get(i - 1).getBoundary();
        Optional<Boundary> b =
            valueBoundary.createBi(BoundaryBiCreaterType.INTERSECTION, missingRuleBoundary);
        if (b.isPresent()) {
          missingRule.setInputValue(
              i - 1,
              VDmnInputValueImpl.getBuilder()
                  .withInputEntryId(
                      new InputEntryId("Entry_" + Long.toString(UniqueIdGenerator.getNextId(), 16)))
                  .withText("")
                  .withDmnInputColumn((VDmnInputColumn) missingRuleValue.getDmnColumn())
                  .withDmnRule(missingRule)
                  .withBoundary(b.get())
                  .build());
        }
      }
      // calc next missing intervals: left + right
      VDmnInputValue currentValue = currentRuleValues.get(i);
      // left
      addMissingRules(
          i,
          inputs,
          missingInValues
              .get(i)
              .getBoundary()
              .createBi(BoundaryBiCreaterType.LOWER_BOUNDS, currentValue.getBoundary()),
          missingRule,
          newMissingRules);

      // right
      addMissingRules(
          i,
          inputs,
          missingInValues
              .get(i)
              .getBoundary()
              .createBi(BoundaryBiCreaterType.UPPER_BOUNDS, currentValue.getBoundary()),
          missingRule,
          newMissingRules);
    }
    return newMissingRules;
  }

  private void addMissingRules(
      int i,
      List<VDmnInputColumn> inputs,
      Optional<Boundary> bi,
      VDmnRuleChangeableImpl missingRule,
      List<VDmnRuleChangeableImpl> newMissingRules) {
    if (bi.isPresent()) {
      List<VDmnInputValue> missingRuleValues = missingRule.getDmnInputValues();

      Builder b =
          VDmnRuleImpl.getBuilder()
              .withRuleId(new RuleId("Rule_" + Long.toString(UniqueIdGenerator.getNextId(), 16)))
              .withDmnDecisionTable(inputs.get(0).getDmnDecisionTable())
              .withRowNumber(new RowNumber(INITIAL_ROW_NUMER));
      for (VDmnInputValue inputValue : missingRuleValues) {
        b.addInputValue(inputValue);
      }

      VDmnRuleChangeableImpl copy = new VDmnRuleChangeableImpl(b.build());

      for (int x = 0; x < missingRuleValues.size(); x++) {
        copy.setInputValue(
            x,
            VDmnInputValueImpl.getBuilder()
                .withInputEntryId(
                    new InputEntryId("Entry_" + Long.toString(UniqueIdGenerator.getNextId(), 16)))
                .withText("")
                .withDmnRule(copy)
                .withBoundary(x == i ? bi.get() : missingRuleValues.get(x).getBoundary())
                .withDmnInputColumn(inputs.get(x))
                .build());
      }
      newMissingRules.add(copy);
    }
  }

  private boolean isInContact(VDmnRule missingRule, VDmnRule currentRule) {
    List<VDmnInputValue> inValsMissingRule = missingRule.getDmnInputValues();
    List<VDmnInputValue> inValsCurrentRule = currentRule.getDmnInputValues();
    for (int i = 0; i < inValsMissingRule.size(); i++) {
      if (inValsMissingRule
          .get(i)
          .getBoundary()
          .checkWith(BoundaryCheckType.IS_NOT_IN_CONTACT, inValsCurrentRule.get(i).getBoundary())) {
        return false;
      }
    }
    return true;
  }

  private void mergeRules(List<VDmnRuleChangeableImpl> missingRules) {
    boolean found = true;
    while (found) {
      List<VDmnRule> removeRules = new ArrayList<>(100);
      found = false;
      for (int i = 0; i < missingRules.size(); i++) {
        List<VDmnInputValue> inputValuesI = missingRules.get(i).getDmnInputValues();
        for (int u = i + 1; u < missingRules.size(); u++) {
          List<VDmnInputValue> inputValuesU = missingRules.get(u).getDmnInputValues();
          int foundCombination = -1;
          Optional<Boundary> newOBoundary;
          Boundary newBoundary = null;

          for (int x = 0; x < inputValuesI.size(); x++) {
            VDmnInputValue v1 = inputValuesI.get(x);
            VDmnInputValue v2 = inputValuesU.get(x);
            newOBoundary =
                v1.getBoundary().createBi(BoundaryBiCreaterType.COMBINE, v2.getBoundary());
            if (v1.getBoundary().checkWith(BoundaryCheckType.IS_EQUAL, v2.getBoundary())) {
              // nothing
            } else if (newOBoundary.isPresent() && foundCombination == -1) {
              newBoundary = newOBoundary.get();
              foundCombination = x;
            } else {
              foundCombination = -1;
              break;
            }
          }
          if (foundCombination >= 0) {
            found = true;
            missingRules
                .get(i)
                .setInputValue(
                    foundCombination,
                    new VDmnInputValueChangeableImpl(
                            (VDmnInputValueImpl) inputValuesI.get(foundCombination))
                        .withNewBoundary(newBoundary));
            removeRules.add(missingRules.get(u));
          }
        }
      }
      missingRules.removeAll(removeRules);
    }
  }
}
