package de.unikoblenz.fgbks.core.dmn.verification.verifier.impl;

import static de.unikoblenz.fgbks.base.utils.boundary.bicreater.BoundaryBiCreaterType.COMBINE;
import static de.unikoblenz.fgbks.base.utils.boundary.checker.BoundaryCheckType.IS_EQUAL;
import static de.unikoblenz.fgbks.base.utils.boundary.checker.BoundaryCheckType.SUBSUMES;
import static de.unikoblenz.fgbks.core.dmn.domain.vdmn.utils.VDmnFunctions.getRulesRowsStrings;
import static de.unikoblenz.fgbks.core.dmn.domain.vdmn.utils.VDmnFunctions.templateDecision;
import static de.unikoblenz.fgbks.core.dmn.verification.result.actions.VerificationFix.SHOW_RULES;

import de.unikoblenz.fgbks.base.domain.Name;
import de.unikoblenz.fgbks.base.utils.boundary.Boundary;
import de.unikoblenz.fgbks.core.dmn.domain.vdmn.VDmnDecision;
import de.unikoblenz.fgbks.core.dmn.domain.vdmn.VDmnDecisionTable;
import de.unikoblenz.fgbks.core.dmn.domain.vdmn.VDmnInputColumn;
import de.unikoblenz.fgbks.core.dmn.domain.vdmn.VDmnInputValue;
import de.unikoblenz.fgbks.core.dmn.domain.vdmn.VDmnRule;
import de.unikoblenz.fgbks.core.dmn.domain.vdmn.utils.VDmnFunctions;
import de.unikoblenz.fgbks.core.dmn.verification.result.VerificationResultEntry.VerificationClassification;
import de.unikoblenz.fgbks.core.dmn.verification.result.VerificationResultEntryElement;
import de.unikoblenz.fgbks.core.dmn.verification.result.actions.Action;
import de.unikoblenz.fgbks.core.dmn.verification.result.actions.ActionScope;
import de.unikoblenz.fgbks.core.dmn.verification.result.actions.ActionType;
import de.unikoblenz.fgbks.core.dmn.verification.result.actions.VerificationFix;
import de.unikoblenz.fgbks.core.dmn.verification.verifier.AbstractVerifier;
import de.unikoblenz.fgbks.core.dmn.verification.verifier.DmnVerifier;
import de.unikoblenz.fgbks.core.dmn.verification.verifier.types.PartialReductionVerification;
import java.util.ArrayList;
import java.util.List;

@DmnVerifier(verifierType = PartialReductionVerification.class)
public class PartialReductionVerifier extends AbstractVerifier {

  @Override
  protected void doVerification() {
    for (VDmnDecision decision : dmnObjectContainer.getVDmnDefinition().getDmnDecisions()) {
      findPartialReduction(decision.getDmnDecisionTable());
    }
  }

  private void findPartialReduction(VDmnDecisionTable dmnDecisionTable) {
    List<List<VDmnRule>> identicalOutputCluster =
        VDmnFunctions.getRuleClustersWithIdenticalOutput(dmnDecisionTable);
    List<VDmnInputColumn> inColumns = dmnDecisionTable.getInputColumns();
    identicalOutputCluster.forEach(c -> findPartialReduction(inColumns, 0, c, false, null));
  }

  private void findPartialReduction(
      List<VDmnInputColumn> inColumns,
      int i,
      List<VDmnRule> clusterRules,
      boolean hasCombination,
      VDmnRule subsumptionRule) {
    if (i == inColumns.size()) {
      if (hasCombination) {
        createCombinedRule(inColumns, clusterRules);
        clusterRules.forEach(
            rule -> vreFactory.addElement(VerificationResultEntryElement.create(rule)));
        vreFactory.addVerificationFix(SHOW_RULES);
        vreFactory.addToEntry(
            VerificationClassification.INFO,
            templateDecision(inColumns.get(0).getDmnDecision()) + "Rules %s can be combined.",
            getRulesRowsStrings(clusterRules));
      }
    } else {
      List<VDmnInputValue> rules =
          VDmnFunctions.getColumnValuesInRules(inColumns.get(i), clusterRules);
      List<List<VDmnRule>> nClusters = new ArrayList<>();
      List<Boolean> combinationCluster = new ArrayList<>();
      for (int i1 = 0; i1 < rules.size(); i1++) {
        VDmnInputValue currentRuleValue = rules.get(i1);
        for (int i2 = i1 + 1; i2 < rules.size(); i2++) {
          VDmnInputValue currentRuleValueInner = rules.get(i2);
          if (currentRuleValue != currentRuleValueInner) {
            boolean found = false;
            // 1. Is subsumption?
            if (currentRuleValue
                .getBoundary()
                .checkWith(SUBSUMES, currentRuleValueInner.getBoundary())) {
              if (subsumptionRule == null || subsumptionRule == currentRuleValue.getDmnRule()) {
                subsumptionRule = currentRuleValue.getDmnRule();
                found = true;
                List<VDmnRule> cluster = new ArrayList<>();
                cluster.add(currentRuleValue.getDmnRule());
                cluster.add(currentRuleValueInner.getDmnRule());
                nClusters.add(cluster);
                combinationCluster.add(true);
              }
            }
            if (currentRuleValueInner
                .getBoundary()
                .checkWith(SUBSUMES, currentRuleValue.getBoundary())) {
              if (subsumptionRule == null
                  || subsumptionRule == currentRuleValueInner.getDmnRule()) {
                subsumptionRule = currentRuleValueInner.getDmnRule();
                found = true;
                List<VDmnRule> cluster = new ArrayList<>();
                cluster.add(currentRuleValue.getDmnRule());
                cluster.add(currentRuleValueInner.getDmnRule());
                nClusters.add(cluster);
                combinationCluster.add(true);
              }
            }
            // 2. Find other combinations (only of no real combination
            if (!found && !hasCombination) {
              if (currentRuleValue
                  .getBoundary()
                  .createBi(COMBINE, currentRuleValueInner.getBoundary())
                  .isPresent()) {
                found = true;
                List<VDmnRule> cluster = new ArrayList<>();
                cluster.add(currentRuleValue.getDmnRule());
                cluster.add(currentRuleValueInner.getDmnRule());
                nClusters.add(cluster);
                combinationCluster.add(true);
              }
            }
            // 2. Find equal
            if (!found
                && currentRuleValue
                .getBoundary()
                .checkWith(IS_EQUAL, currentRuleValueInner.getBoundary())) {
              List<VDmnRule> cluster = new ArrayList<>();
              cluster.add(currentRuleValue.getDmnRule());
              cluster.add(currentRuleValueInner.getDmnRule());
              nClusters.add(cluster);
              combinationCluster.add(false);
            }
          }
        }
      }
      // nex col
      for (int x = 0; x < combinationCluster.size(); x++) {
        if (nClusters.get(x).size() > 1) {
          findPartialReduction(
              inColumns,
              i + 1,
              nClusters.get(x),
              combinationCluster.get(x) || hasCombination,
              subsumptionRule);
        }
      }
    }
  }

  private void createCombinedRule(List<VDmnInputColumn> inColumns, List<VDmnRule> clusterRules) {
    // Combined Rule:
    List<Boundary> boundaries = new ArrayList<>();
    for (int i = 0; i < inColumns.size(); i++) {
      boundaries.add(
          clusterRules
              .get(0)
              .getDmnInputValues()
              .get(i)
              .getBoundary()
              .createBi(COMBINE, clusterRules.get(1).getDmnInputValues().get(i).getBoundary())
              .get());
    }
    Action.Builder fixActionBuilder =
        Action.getBuilder().withActionScope(ActionScope.RULE).withActionType(ActionType.CREATE);
    for (int i = 0; i < inColumns.size(); i++) {
      fixActionBuilder.addValue(
          inColumns.get(i).getInputId().getValue(), boundaries.get(i).getParsedText());
    }
    Action.Builder fixActionBuilder2 =
        Action.getBuilder().withActionScope(ActionScope.RULE).withActionType(ActionType.DELETE);
    fixActionBuilder2.addValue(
        "decisionId", clusterRules.get(0).getDmnDecision().getDecisionId().getValue());
    fixActionBuilder2.addValue("ruleId", clusterRules.get(0).getRuleId().getValue());
    Action.Builder fixActionBuilder3 =
        Action.getBuilder().withActionScope(ActionScope.RULE).withActionType(ActionType.DELETE);
    fixActionBuilder3.addValue(
        "decisionId", clusterRules.get(1).getDmnDecision().getDecisionId().getValue());
    fixActionBuilder3.addValue("ruleId", clusterRules.get(1).getRuleId().getValue());
    vreFactory.addVerificationFix(
        VerificationFix.getBuilder()
            .withFixName(new Name("Combine"))
            .addAction(fixActionBuilder.build())
            .addAction(fixActionBuilder2.build())
            .addAction(fixActionBuilder3.build())
            .build());
  }
}
